package xyz.lychee.lagfixer.libs.adventure.text;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style.Merge.Strategy;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.annotations.VisibleForTesting;

final class ComponentCompaction {
   @VisibleForTesting
   static final boolean SIMPLIFY_STYLE_FOR_BLANK_COMPONENTS = false;

   private ComponentCompaction() {
   }

   static Component compact(final @NotNull Component self, final @Nullable Style parentStyle) {
      List<Component> children = self.children();
      Component optimized = self.children(Collections.emptyList());
      if (parentStyle != null) {
         optimized = ((Component)optimized).style(self.style().unmerge(parentStyle));
      }

      int childrenSize = children.size();
      if (childrenSize == 0) {
         if (isBlank((Component)optimized)) {
            optimized = ((Component)optimized).style(simplifyStyleForBlank(((Component)optimized).style(), parentStyle));
         }

         return (Component)optimized;
      } else {
         if (childrenSize == 1 && optimized instanceof TextComponent) {
            TextComponent textComponent = (TextComponent)optimized;
            if (textComponent.content().isEmpty()) {
               Component child = (Component)children.get(0);
               return child.style(child.style().merge(((Component)optimized).style(), Strategy.IF_ABSENT_ON_TARGET)).compact();
            }
         }

         Style childParentStyle = ((Component)optimized).style();
         if (parentStyle != null) {
            childParentStyle = childParentStyle.merge(parentStyle, Strategy.IF_ABSENT_ON_TARGET);
         }

         List<Component> childrenToAppend = new ArrayList(children.size());

         int i;
         Component child;
         for(i = 0; i < children.size(); ++i) {
            child = (Component)children.get(i);
            child = compact(child, childParentStyle);
            if (child.children().isEmpty() && child instanceof TextComponent) {
               TextComponent textComponent = (TextComponent)child;
               if (textComponent.content().isEmpty()) {
                  continue;
               }
            }

            childrenToAppend.add(child);
         }

         if (optimized instanceof TextComponent) {
            while(!childrenToAppend.isEmpty()) {
               Component child = (Component)childrenToAppend.get(0);
               Style childStyle = child.style().merge(childParentStyle, Strategy.IF_ABSENT_ON_TARGET);
               if (!(child instanceof TextComponent) || !Objects.equals(childStyle, childParentStyle)) {
                  break;
               }

               optimized = joinText((TextComponent)optimized, (TextComponent)child);
               childrenToAppend.remove(0);
               childrenToAppend.addAll(0, child.children());
            }
         }

         i = 0;

         while(true) {
            while(i + 1 < childrenToAppend.size()) {
               child = (Component)childrenToAppend.get(i);
               Component neighbor = (Component)childrenToAppend.get(i + 1);
               if (child.children().isEmpty() && child instanceof TextComponent && neighbor instanceof TextComponent) {
                  Style childStyle = child.style().merge(childParentStyle, Strategy.IF_ABSENT_ON_TARGET);
                  Style neighborStyle = neighbor.style().merge(childParentStyle, Strategy.IF_ABSENT_ON_TARGET);
                  if (childStyle.equals(neighborStyle)) {
                     Component combined = joinText((TextComponent)child, (TextComponent)neighbor);
                     childrenToAppend.set(i, combined);
                     childrenToAppend.remove(i + 1);
                     continue;
                  }
               }

               ++i;
            }

            if (childrenToAppend.isEmpty() && isBlank((Component)optimized)) {
               optimized = ((Component)optimized).style(simplifyStyleForBlank(((Component)optimized).style(), parentStyle));
            }

            return ((Component)optimized).children(childrenToAppend);
         }
      }
   }

   private static boolean isBlank(final Component component) {
      if (component instanceof TextComponent) {
         TextComponent textComponent = (TextComponent)component;
         String content = textComponent.content();

         for(int i = 0; i < content.length(); ++i) {
            char c = content.charAt(i);
            if (c != ' ') {
               return false;
            }
         }

         return true;
      } else {
         return false;
      }
   }

   private static @NotNull Style simplifyStyleForBlank(final @NotNull Style style, final @Nullable Style parentStyle) {
      return style;
   }

   private static TextComponent joinText(final TextComponent one, final TextComponent two) {
      return TextComponentImpl.create(two.children(), one.style(), one.content() + two.content());
   }
}
